import { StringFilterTypes, type AnyStatement } from 'ts-common/src/model/query/logic/general';
import { log } from 'ts-common/src/logger/logger';
import type { NodeStruct } from 'ts-common';

export type SQLCacheData = {
  nodes: NodeStruct[];
};

export function createWhereLogic(
  op: string,
  left: string,
  whereLogic: Set<string>,
  cacheData: SQLCacheData,
): { logic: string; where: Set<string> } {
  const newWhereLogic = `${left.replace('.', '_')}_${op}`;
  const tableName = left.split('.')[0];

  whereLogic.add(`${op}(${left}) AS ${newWhereLogic}`);
  for (const entity of cacheData.nodes) {
    if (entity.id !== tableName && entity.id) {
      whereLogic.add(entity.id);
    }
  }

  return { logic: newWhereLogic, where: whereLogic };
}

export function extractLogicSQL(logicQuery: AnyStatement | AnyStatement[], cacheData: SQLCacheData): { logic: string; where: Set<string> } {
  switch (typeof logicQuery) {
    case 'object':
      if (Array.isArray(logicQuery)) {
        logicQuery = logicQuery as AnyStatement[];
        let op = (logicQuery[0] as string).replace('_', '').toLowerCase();
        const right = logicQuery?.[2];
        const { logic: left, where: whereLogic } = extractLogicSQL(logicQuery[1], cacheData);

        if (left === '') {
          return { logic: '', where: whereLogic };
        }

        switch (op) {
          case '!=':
            op = '!=';
            break;
          case '==':
            op = '=';
            break;
          case 'like':
            return {
              logic: `${left} LIKE '%${(right as string).replace(/['"]/g, '')}%'`,
              where: whereLogic,
            };
          case 'in':
          case 'present in list':
            if (typeof right === 'string') {
              return {
                logic: `${left} IN (${right
                  .replace(/"/g, "'")
                  .split(',')
                  .map((r: string) => r.trim())
                  .join(', ')})`,
                where: whereLogic,
              };
            }
            break;
          case 'not in':
          case 'not in list':
            if (typeof right === 'string') {
              return {
                logic: `${left} NOT IN (${right
                  .replace(/"/g, "'")
                  .split(',')
                  .map((r: string) => r.trim())
                  .join(', ')})`,
                where: whereLogic,
              };
            }
            break;
          case 'isempty':
            return { logic: `(${left} IS NULL OR ${left} = '')`, where: whereLogic };
          case 'isnotempty':
            return { logic: `(${left} IS NOT NULL AND ${left} <> '')`, where: whereLogic };
          case 'lower':
            return { logic: `LOWER(${left})`, where: whereLogic };
          case 'upper':
            return { logic: `UPPER(${left})`, where: whereLogic };
          case 'avg':
            return {
              logic: `(SELECT AVG(${left}) FROM ${left.split('.')[0]})`,
              where: whereLogic,
            };
          case 'count':
            return {
              logic: `(SELECT COUNT(${left}) FROM ${left.split('.')[0]})`,
              where: whereLogic,
            };
          case 'max':
            return {
              logic: `(SELECT MAX(${left}) FROM ${left.split('.')[0]})`,
              where: whereLogic,
            };
          case 'min':
            return {
              logic: `(SELECT MIN(${left}) FROM ${left.split('.')[0]})`,
              where: whereLogic,
            };
          case 'sum':
            return {
              logic: `(SELECT SUM(${left}) FROM ${left.split('.')[0]})`,
              where: whereLogic,
            };
          case 'and':
            if (logicQuery.length > 2) {
              const conditions = logicQuery.slice(1).map(condition => {
                const { logic: condLogic } = extractLogicSQL(condition, cacheData);
                return condLogic;
              });
              return { logic: `(${conditions.join(' AND ')})`, where: whereLogic };
            }
            break;
          case 'or':
            if (logicQuery.length > 2) {
              const conditions = logicQuery.slice(1).map(condition => {
                const { logic: condLogic } = extractLogicSQL(condition, cacheData);
                return condLogic;
              });
              return { logic: `(${conditions.join(' OR ')})`, where: whereLogic };
            }
            break;
        }
        if (logicQuery.length > 2 && logicQuery[2] !== undefined) {
          const { logic: rightLogic, where: whereLogicRight } = extractLogicSQL(logicQuery[2], cacheData);

          for (const where of whereLogicRight) {
            whereLogic.add(where);
          }

          return { logic: `(${left} ${op} ${rightLogic})`, where: whereLogic };
        }
        return { logic: `(${op} ${left})`, where: whereLogic };
      }
      return { logic: logicQuery, where: new Set() };
    case 'string':
      // Handle string literals and field references
      if (logicQuery.startsWith('"') && logicQuery.endsWith('"')) {
        return { logic: logicQuery.replace(/"/g, "'"), where: new Set() };
      }
      if (logicQuery.startsWith('@')) {
        return { logic: logicQuery.replace('@', ''), where: new Set() };
      } else {
        return { logic: `'${logicQuery}'`, where: new Set() };
      }
    case 'number':
      return { logic: logicQuery.toString(), where: new Set() };
    default:
      return { logic: logicQuery as any, where: new Set() };
  }
}
