import type { BackendQueryFormat, NodeStruct } from 'ts-common';
import type { QueryText } from '../model';

type Logic = any;

function cleanStringOfNewLinesAndTabs(str: string): string {
  if (!str) {
    return '';
  }
  let trimmedSQL = str.replace(/\n/g, ' ');
  trimmedSQL = trimmedSQL.replaceAll(/ {2,50}/g, ' ');
  trimmedSQL = trimmedSQL.replace(/\t+/g, '');
  return trimmedSQL.trim();
}

function convertLogicToSQL(logic: Logic, nodeAlias: string): string {
  if (!logic || !Array.isArray(logic)) return '';

  const operator = logic[0];
  const left = logic[1];
  const right = logic[2];

  // Handle field references (e.g., @p1.name)
  if (typeof left === 'string' && left.startsWith('@')) {
    const [alias, field] = left.substring(1).split('.');
    if (alias === nodeAlias) {
      return `${alias}.${field} ${operator} '${right.replace(/^"(.*)"$/, '$1')}'`;
    }
  }
  return '';
}

function buildNodeCTE(node: NodeStruct | undefined, logic?: Logic, parent: NodeStruct | undefined = undefined): string[] {
  if (!node || !node.label || !node.id) {
    return [];
  }
  const nodeParts: string[] = [];
  const logicClause = convertLogicToSQL(logic, node.id);
  const whereClause = logic && logicClause ? `WHERE ${convertLogicToSQL(logic, node.id)}` : '';
  const mainNode = `
  SELECT jsonb_build_object(
    '_id', concat('${node.label}-', ${node.id}.id::text),
    'label', '${node.label}',
    'attributes', to_jsonb(${node.id}) - 'id'
  ) AS node
  FROM ${node.label} ${node.id}
  ${
    parent && parent.label && parent.id && parent.relation && parent.relation.label
      ? `INNER JOIN ${parent.label} ${parent.id} ON ${parent.id}.${parent.relation.label.split('.')[0]} = ${node.id}.${
          parent.relation.label.split('.')[1]
        }`
      : ''
  }
  ${whereClause}`;
  nodeParts.push(cleanStringOfNewLinesAndTabs(mainNode));
  buildNodeCTE(node.relation?.node, logic, node).forEach(part => nodeParts.push(part));
  return nodeParts;
}

function buildEdgesCTE(node: NodeStruct | undefined, logic?: Logic, parent: NodeStruct | undefined = undefined): string[] {
  if (!node || !node.relation || !node.relation.node || !node.id || !node.relation.node.id || !node.relation.label) {
    return [];
  }
  const logicClause = convertLogicToSQL(logic, node.id);
  const whereClause = logic && logicClause ? `WHERE ${convertLogicToSQL(logic, node.id)}` : '';

  const edge = `
  SELECT jsonb_agg(
    jsonb_build_object(
      '_id', concat('${node.label}-', ${node.id}.id::text, '-${node.relation.node.label}-', ${node.relation.node.id}.id::text),
      'label', '${node.relation.label}',
      'attributes', '{}'::jsonb,
      'from', ${node.id}.id::text,
      'to', ${node.relation.node.id}.id::text
    )
  ) AS edge
  FROM ${node.label} ${node.id}
  INNER JOIN ${node.relation.node.label} ${node.relation.node.id} ON ${node.id}.${node.relation.label.split('.')[0]} = ${
    node.relation.node.id
  }.${node.relation.label.split('.')[1]}
  ${whereClause}`;
  const ret = cleanStringOfNewLinesAndTabs(edge);
  return [ret, ...buildEdgesCTE(node.relation.node, logic, node)];
}

function buildNodeCountsCTE(node: NodeStruct | undefined): string[] {
  if (!node || !node.label || !node.id) {
    return [];
  }
  const nodeParts: string[] = [];
  const mainNode = `SELECT '${node.label}' AS label FROM ${node.label}`;
  nodeParts.push(cleanStringOfNewLinesAndTabs(mainNode));
  buildNodeCountsCTE(node.relation?.node).forEach(part => nodeParts.push(part));
  return nodeParts;
}

function buildEdgeCountsCTE(node: NodeStruct | undefined): string[] {
  if (!node || !node.relation || !node.relation.node || !node.id || !node.relation.node.id || !node.relation.label) {
    return [];
  }

  const edge = `
  SELECT '${node.relation.label}' AS label, COUNT(*) AS count
  FROM ${node.label} ${node.id}
  JOIN ${node.relation.node.label} ${node.relation.node.id}
  ON ${node.id}.${node.relation.label.split('.')[0]} = ${node.relation.node.id}.${node.relation.label.split('.')[1]}
  GROUP BY label`;
  const ret = cleanStringOfNewLinesAndTabs(edge);
  return [ret, ...buildEdgeCountsCTE(node.relation.node)];
}

export function query2SQL(query: BackendQueryFormat): QueryText {
  const nodesCTE = [...new Set(query.query.flatMap(path => buildNodeCTE(path.node, query.logic)))];
  const edgesCTE = [...new Set(query.query.flatMap(path => buildEdgesCTE(path.node)))];
  const nodeCountsCTE = [...new Set(query.query.flatMap(path => buildNodeCountsCTE(path.node)))];
  const edgeCountsCTE = [...new Set(query.query.flatMap(path => buildEdgeCountsCTE(path.node)))];

  const mainQuery = `WITH raw_nodes AS (
      SELECT node
      FROM (
        ${nodesCTE.join(' UNION ALL ')}
      ) sub
      ${query.limit ? `LIMIT ${query.limit}` : ''}
    ),
    nodes_cte AS (
      SELECT jsonb_agg(node) AS nodes
      FROM raw_nodes
    ),
    edges_cte AS (
    SELECT jsonb_agg(edge) AS edges
      FROM (
        ${edgesCTE.join(' UNION ALL ')}
      ) aggregated_edges
    ),
    node_counts AS (
      SELECT label, COUNT(*) AS count
      FROM (
        ${nodeCountsCTE.join(' UNION ALL ')}
      ) t
      GROUP BY label
    ),
    edge_counts AS (
      ${edgeCountsCTE.join(' UNION ALL ')}
    )
    SELECT jsonb_build_object(
      'nodes', nodes_cte.nodes,
      'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
      'nodeCounts', COALESCE( (SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                               FROM node_counts), '[]'::jsonb ),
      'edgeCounts', COALESCE( (SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                               FROM edge_counts), '[]'::jsonb )
    ) AS result FROM nodes_cte, edges_cte;`;

  return {
    query: mainQuery,
    countQuery: '',
  };
}
