import type { BackendQueryFormat } from 'ts-common';
import { expect, test, describe, it } from 'bun:test';
import { query2SQL } from './queryConverterSql';

function fixSQLSpaces(sql?: string | null): string {
  if (!sql) {
    return '';
  }
  let trimmedSQL = sql.replace(/\n/g, ' ');
  trimmedSQL = trimmedSQL.replaceAll(/ {2,50}/g, ' ');
  trimmedSQL = trimmedSQL.replace(/\t+/g, '');
  return trimmedSQL.trim();
}

describe('query2SQL', () => {
  it('should return correctly on a simple query with multiple paths', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      return: ['*'],
      query: [
        {
          id: 'path1',
          node: {
            label: 'user_management_service_savestate',
            id: 'ss',
            relation: {
              label: 'user_id.id',
              direction: 'TO',
              depth: { min: 1, max: 1 },
              node: {
                label: 'user_management_service_user',
                id: 'uu',
              },
            },
          },
        },
        {
          id: 'path2',
          node: {
            label: 'user_management_service_savestate',
            id: 'ss',
            relation: {
              label: 'id.save_state_id',
              direction: 'TO',
              depth: { min: 1, max: 1 },
              node: {
                label: 'user_management_service_querystate',
                id: 'qs',
              },
            },
          },
        },
      ],
      limit: 5000,
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH raw_nodes AS (
    SELECT node
    FROM (
        SELECT jsonb_build_object(
            '_id', concat('user_management_service_savestate-', ss.id::text),
            'label', 'user_management_service_savestate',
            'attributes', to_jsonb(ss) - 'id'
        ) AS node
        FROM user_management_service_savestate ss
        UNION ALL
        SELECT jsonb_build_object(
            '_id', concat('user_management_service_user-', uu.id::text),
            'label', 'user_management_service_user',
            'attributes', to_jsonb(uu) - 'id'
        ) AS node
        FROM user_management_service_user uu
        UNION ALL
        SELECT jsonb_build_object(
            '_id', concat('user_management_service_querystate-', qs.id::text),
            'label', 'user_management_service_querystate',
            'attributes', to_jsonb(qs) - 'id'
        ) AS node
        FROM user_management_service_querystate qs
    ) sub
    LIMIT 5000
),
nodes_cte AS (
    SELECT jsonb_agg(node) AS nodes
    FROM raw_nodes
),
edges_cte AS (
    SELECT jsonb_agg(edge) AS edges
    FROM ( SELECT jsonb_agg(
        jsonb_build_object(
            '_id', concat('user_management_service_savestate-', ss.id::text, '-user_management_service_user-', uu.id::text),
            'label', 'user_id.id',
            'attributes', '{}'::jsonb,
            'from', ss.id::text,
            'to', uu.id::text
        )
      ) AS edge
      FROM user_management_service_savestate ss
      JOIN user_management_service_user uu
          ON ss.user_id = uu.id
          
      UNION ALL
      
      SELECT jsonb_agg(
          jsonb_build_object(
              '_id', concat('user_management_service_savestate-', ss.id::text, '-user_management_service_querystate-', qs.id::text),
              'label', 'id.save_state_id',
              'attributes', '{}'::jsonb,
              'from', ss.id::text,
              'to', qs.id::text
          )
      ) AS edge
      FROM user_management_service_savestate ss
      JOIN user_management_service_querystate qs
          ON ss.id = qs.save_state_id
    ) aggregated_edges
),
node_counts AS (
    SELECT label, COUNT(*) AS count
    FROM (
        SELECT 'user_management_service_savestate' AS label
        FROM user_management_service_savestate
        UNION ALL
        SELECT 'user_management_service_user' AS label
        FROM user_management_service_user
        UNION ALL
        SELECT 'user_management_service_querystate' AS label
        FROM user_management_service_querystate
    ) t
    GROUP BY label
),
edge_counts AS (
    SELECT 'user_id.id' AS label, COUNT(*) AS count
    FROM user_management_service_savestate ss
    JOIN user_management_service_user uu
        ON ss.user_id = uu.id
    GROUP BY label
    UNION ALL
    SELECT 'id.save_state_id' AS label, COUNT(*) AS count
    FROM user_management_service_savestate ss
    JOIN user_management_service_querystate qs
        ON ss.id = qs.save_state_id
    GROUP BY label
)
SELECT jsonb_build_object(
    'nodes', nodes_cte.nodes,
    'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
    'nodeCounts', COALESCE(
        (SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
         FROM node_counts), 
        '[]'::jsonb
    ),
    'edgeCounts', COALESCE(
        (SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
         FROM edge_counts), 
        '[]'::jsonb
    )
) AS result
FROM nodes_cte, edges_cte;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });

  it('should return correctly on a complex query with logic', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      return: ['*'],
      logic: ['!=', '@uu.username', '"Username1"'],
      query: [
        {
          id: 'path1',
          node: {
            label: 'user_management_service_savestate',
            id: 'ss',
            relation: {
              label: 'user_id.id',
              direction: 'TO',
              depth: { min: 1, max: 1 },
              node: {
                label: 'user_management_service_user',
                id: 'uu',
              },
            },
          },
        },
        {
          id: 'path2',
          node: {
            label: 'user_management_service_savestate',
            id: 'ss',
            relation: {
              label: 'id.save_state_id',
              direction: 'TO',
              depth: { min: 1, max: 1 },
              node: {
                label: 'user_management_service_querystate',
                id: 'qs',
              },
            },
          },
        },
      ],
      limit: 5000,
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH nodes_cte AS (
  SELECT jsonb_agg(node) AS nodes
  FROM (
    SELECT jsonb_build_object(
      '_id', concat('Person-', p1.id::text),
      'label', 'Person',
      'attributes', to_jsonb(p1) - 'id'
    ) AS node
    FROM Person p1
    UNION ALL
    SELECT jsonb_build_object(
      '_id', concat('Movie-', m1.id::text),
      'label', 'Movie',
      'attributes', to_jsonb(m1) - 'id'
    ) AS node
    FROM Movie m1
    UNION ALL
    SELECT jsonb_build_object(
      '_id', concat('Genre-', g1.id::text),
      'label', 'Genre',
      'attributes', to_jsonb(g1) - 'id'
    ) AS node
    FROM Genre g1
  ) sub
),
edges_cte AS (
  SELECT jsonb_agg(
    jsonb_build_object(
      '_id', concat('Person-', p1.id::text, '-Movie-', m1.id::text),
      'label', 'watched.id',
      'attributes', '{}'::jsonb,
      'from', p1.id::text,
      'to', m1.id::text
    )
  ) AS edges
  FROM Person p1
  JOIN Movie m1 ON p1.watched = m1.id
  UNION ALL
  SELECT jsonb_agg(
    jsonb_build_object(
      '_id', concat('Person-', p1.id::text, '-Genre-', g1.id::text),
      'label', 'watched_genre.id',
      'attributes', '{}'::jsonb,
      'from', p1.id::text,
      'to', g1.id::text
    )
  ) AS edges
  FROM Person p1
  JOIN Genre g1 ON p1.watched_genre = g1.id
),
node_counts AS (
  SELECT label, COUNT(*) AS count
  FROM (
    SELECT 'Person' AS label FROM Person
    UNION ALL
    SELECT 'Movie' AS label FROM Movie
    UNION ALL
    SELECT 'Genre' AS label FROM Genre
  ) t
  GROUP BY label
),
edge_counts AS (
  SELECT 'watched.id' AS label, COUNT(*) AS count FROM Person p1 JOIN Movie m1 ON p1.watched = m1.id GROUP BY label
  UNION ALL
  SELECT 'watched_genre.id' AS label, COUNT(*) AS count FROM Person p1 JOIN Genre g1 ON p1.watched_genre = g1.id GROUP BY label
)
SELECT jsonb_build_object(
  'nodes', nodes_cte.nodes,
  'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
  'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM node_counts), '[]'::jsonb),
  'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM edge_counts), '[]'::jsonb)
) AS result
      WHERE p1.name != 'Raymond Campbell'
      LIMIT 5000;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });

  //   it('should return correctly on a query with group by logic', () => {
  //     const query: BackendQueryFormat = {
  //       saveStateID: 'test',
  //       limit: 5000,
  //       logic: ['And', ['<', '@movie.imdbRating', 7.5], ['!=', 'p2.age', 'p1.age']],
  //       query: [
  //         {
  //           id: 'path1',
  //           node: {
  //             label: 'Person',
  //             id: 'p1',
  //             relation: {
  //               id: 'doesnotmatter',
  //               label: 'acted.name',
  //               depth: { min: 1, max: 1 },
  //               direction: 'TO',
  //               node: {
  //                 label: 'Movie',
  //                 id: 'movie',
  //               },
  //             },
  //           },
  //         },
  //         {
  //           id: 'path2',
  //           node: {
  //             label: 'Person',
  //             id: 'p2',
  //           },
  //         },
  //       ],
  //       return: ['@path2'],
  //     };

  //     const sql = query2SQL(query);
  //     const expectedSQL = `SELECT p2.* FROM Person p1
  //       JOIN Movie movie ON p1.acted = movie.name
  //       JOIN Person AS p2 ON p2.id IS NOT NULL
  //       WHERE movie.imdbRating < 7.5 AND p2.age = p1.age
  //       LIMIT 5000;`;

  //     expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  //   });

  it('should return correctly on a query with no label', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      limit: 5000,
      logic: ['<', ['-', '@movie.year', '@p1.year'], 10],
      query: [
        {
          id: 'path1',
          node: {
            id: 'p1',
            filter: [],
            relation: {
              id: 'asdasd',
              label: 'acted.id',
              depth: { min: 1, max: 1 },
              direction: 'TO',
              node: {
                label: 'Movie',
                id: 'movie',
              },
            },
          },
        },
      ],
      return: ['*'],
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH nodes_cte AS (
  SELECT jsonb_agg(node) AS nodes
  FROM (
    -- Nodes from p1
    SELECT jsonb_build_object(
      '_id', concat('p1-', p1.id::text),
      'label', 'p1',
      'attributes', to_jsonb(p1) - 'id'
    ) AS node
    FROM p1
    UNION ALL
    -- Nodes from Movie
    SELECT jsonb_build_object(
      '_id', concat('Movie-', movie.id::text),
      'label', 'Movie',
      'attributes', to_jsonb(movie) - 'id'
    ) AS node
    FROM Movie movie
  ) sub
),
edges_cte AS (
  SELECT jsonb_agg(
    jsonb_build_object(
      '_id', concat('p1-', p1.id::text, '-Movie-', movie.id::text),
      'label', 'acted.id',
      'attributes', '{}'::jsonb,
      'from', p1.id::text,
      'to', movie.id::text
    )
  ) AS edges
  FROM p1
  JOIN Movie movie ON p1.acted = movie.id
),
node_counts AS (
  SELECT label, COUNT(*) AS count
  FROM (
    SELECT 'p1' AS label FROM p1
    UNION ALL
    SELECT 'Movie' AS label FROM Movie
  ) t
  GROUP BY label
),
edge_counts AS (
  SELECT 'acted.id' AS label, COUNT(*) AS count FROM p1 JOIN Movie movie ON p1.acted = movie.id GROUP BY label
)
SELECT jsonb_build_object(
  'nodes', nodes_cte.nodes,
  'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
  'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM node_counts), '[]'::jsonb),
  'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM edge_counts), '[]'::jsonb)
) AS result
      WHERE (movie.year - p1.year) < 10
      LIMIT 5000;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });

  //   it('should return correctly on a query with no depth', () => {
  //     const query: BackendQueryFormat = {
  //       saveStateID: 'test',
  //       limit: 5000,
  //       logic: ['And', ['<', '@movie.imdbRating', 7.5], ['==', 'p2.age', 'p1.age']],
  //       query: [
  //         {
  //           id: 'path1',
  //           node: {
  //             id: 'p1',
  //             relation: {
  //               id: 'acted',
  //               direction: 'TO',
  //               node: {
  //                 label: 'Movie',
  //                 id: 'movie',
  //               },
  //             },
  //           },
  //         },
  //         {
  //           id: 'path2',
  //           node: {
  //             id: 'p2',
  //           },
  //         },
  //       ],
  //       return: ['*'],
  //     };

  //     const sql = query2SQL(query);
  //     const expectedSQL = `SELECT * FROM p1
  //       JOIN Movie movie ON p1.id = movie.id AND movie.relation = 'acted'
  //       JOIN p2 ON /* join condition assumed */
  //       WHERE movie.imdbRating < 7.5 AND p2.age = p1.age
  //       LIMIT 5000;`;

  //     expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  //   });

  it('should return correctly on a query with average calculation', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      limit: 5000,
      logic: ['<', '@p1.age', ['Avg', '@p1.age']],
      query: [
        {
          id: 'path1',
          node: {
            label: 'Person',
            id: 'p1',
            relation: {
              id: 'acted',
              label: 'acted.id',
              depth: { min: 1, max: 1 },
              direction: 'TO',
              node: {
                label: 'Movie',
                id: 'movie',
              },
            },
          },
        },
      ],
      return: ['*'],
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH nodes_cte AS (
  SELECT jsonb_agg(node) AS nodes
  FROM (
    -- Nodes from Person
    SELECT jsonb_build_object(
      '_id', concat('Person-', p1.id::text),
      'label', 'Person',
      'attributes', to_jsonb(p1) - 'id'
    ) AS node
    FROM Person p1
    UNION ALL
    -- Nodes from Movie
    SELECT jsonb_build_object(
      '_id', concat('Movie-', movie.id::text),
      'label', 'Movie',
      'attributes', to_jsonb(movie) - 'id'
    ) AS node
    FROM Movie movie
  ) sub
),
edges_cte AS (
  SELECT jsonb_agg(
    jsonb_build_object(
      '_id', concat('Person-', p1.id::text, '-Movie-', movie.id::text),
      'label', 'acted.id',
      'attributes', '{}'::jsonb,
      'from', p1.id::text,
      'to', movie.id::text
    )
  ) AS edges
  FROM Person p1
  JOIN Movie movie ON p1.acted = movie.id
),
node_counts AS (
  SELECT label, COUNT(*) AS count
  FROM (
    SELECT 'Person' AS label FROM Person
    UNION ALL
    SELECT 'Movie' AS label FROM Movie
  ) t
  GROUP BY label
),
edge_counts AS (
  SELECT 'acted.id' AS label, COUNT(*) AS count FROM Person p1 JOIN Movie movie ON p1.acted = movie.id GROUP BY label
)
SELECT jsonb_build_object(
  'nodes', nodes_cte.nodes,
  'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
  'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM node_counts), '[]'::jsonb),
  'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM edge_counts), '[]'::jsonb)
) AS result
      WHERE p1.age < (SELECT AVG(p1.age) FROM Person p1 JOIN Movie movie ON p1.acted = movie.id)
      LIMIT 5000;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });

  it('should return correctly on a query with average calculation and multiple paths', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      limit: 5000,
      logic: ['<', '@p1.age', ['Avg', '@p1.age']],
      query: [
        {
          id: 'path1',
          node: {
            label: 'Person',
            id: 'p1',
            relation: {
              id: 'someid',
              label: 'acted.id',
              depth: { min: 1, max: 1 },
              direction: 'TO',
              node: {
                label: 'Movie',
                id: 'movie',
              },
            },
          },
        },
        {
          id: 'path2',
          node: {
            label: 'Person',
            id: 'p2',
            relation: {
              id: 'acted',
              label: 'acted.id',
              depth: { min: 1, max: 1 },
              direction: 'TO',
              node: {
                label: 'Movie',
                id: 'movie',
              },
            },
          },
        },
      ],
      return: ['*'],
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH nodes_cte AS (
  SELECT jsonb_agg(node) AS nodes
  FROM (
    -- Nodes from Person
    SELECT jsonb_build_object(
      '_id', concat('Person-', p1.id::text),
      'label', 'Person',
      'attributes', to_jsonb(p1) - 'id'
    ) AS node
    FROM Person p1
    UNION ALL
    -- Nodes from Movie
    SELECT jsonb_build_object(
      '_id', concat('Movie-', movie.id::text),
      'label', 'Movie',
      'attributes', to_jsonb(movie) - 'id'
    ) AS node
    FROM Movie movie
    UNION ALL
    -- Nodes from Person
    SELECT jsonb_build_object(
      '_id', concat('Person-', p2.id::text),
      'label', 'Person',
      'attributes', to_jsonb(p2) - 'id'
    ) AS node
    FROM Person p2
  ) sub
),
edges_cte AS (
  SELECT jsonb_agg(
    jsonb_build_object(
      '_id', concat('Person-', p1.id::text, '-Movie-', movie.id::text),
      'label', 'acted.id',
      'attributes', '{}'::jsonb,
      'from', p1.id::text,
      'to', movie.id::text
    )
  ) AS edges
  FROM Person p1
  JOIN Movie movie ON p1.id = movie.id AND movie.relation = 'ACTED_IN'
  UNION ALL
  SELECT jsonb_agg(
    jsonb_build_object(
      '_id', concat('Person-', p2.id::text, '-Movie-', movie.id::text),
      'label', 'acted.id',
      'attributes', '{}'::jsonb,
      'from', p2.id::text,
      'to', movie.id::text
    )
  ) AS edges
  FROM Person p2
  JOIN Movie movie ON p2.id = movie.id AND movie.relation = 'ACTED_IN'
),
node_counts AS (
  SELECT label, COUNT(*) AS count
  FROM (
    SELECT 'Person' AS label FROM Person
    UNION ALL
    SELECT 'Movie' AS label FROM Movie
  ) t
  GROUP BY label
),
edge_counts AS (
  SELECT 'acted.id' AS label, COUNT(*) AS count FROM Person p1 JOIN Movie movie ON p1.id = movie.id AND movie.relation = 'ACTED_IN' GROUP BY label
  UNION ALL
    SELECT 'acted.id' AS label, COUNT(*) AS count FROM Person p2 JOIN Movie movie ON p2.id = movie.id AND movie.relation = 'ACTED_IN' GROUP BY label
)
SELECT jsonb_build_object(
  'nodes', nodes_cte.nodes,
  'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
  'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM node_counts), '[]'::jsonb),
  'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM edge_counts), '[]'::jsonb)
) AS result
      WHERE p1.age < (SELECT AVG(p1.age) FROM Person p1 JOIN Movie movie ON p1.id = movie.id AND movie.relation = 'ACTED_IN')
      LIMIT 5000;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });

  it('should return correctly on a single entity query with lower like logic', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      limit: 5000,
      logic: ['Like', ['Lower', '@p1.name'], '"john"'],
      query: [
        {
          id: 'path1',
          node: {
            label: 'Person',
            id: 'p1',
          },
        },
      ],
      return: ['*'],
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH nodes_cte AS (
  SELECT jsonb_agg(node) AS nodes
  FROM (
    -- Nodes from Person
    SELECT jsonb_build_object(
      '_id', concat('Person-', p1.id::text),
      'label', 'Person',
      'attributes', to_jsonb(p1) - 'id'
    ) AS node
    FROM Person p1
  ) sub
),
edges_cte AS (
  SELECT '[]'::jsonb AS edges
),
node_counts AS (
  SELECT label, COUNT(*) AS count
  FROM (
    SELECT 'Person' AS label FROM Person
  ) t
  GROUP BY label
),
edge_counts AS (
    SELECT null AS label, 0 AS count
)
SELECT jsonb_build_object(
  'nodes', nodes_cte.nodes,
  'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
  'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM node_counts), '[]'::jsonb),
  'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM edge_counts), '[]'::jsonb)
) AS result
      WHERE LOWER(p1.name) LIKE '%john%'
      LIMIT 5000;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });

  it('should return correctly on a query with like logic', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      limit: 500,
      logic: ['Like', '@id_1691576718400.title', '"ale"'],
      query: [
        {
          id: 'path_0',
          node: {
            id: 'id_1691576718400',
            label: 'Employee',
            relation: {
              id: 'id_1691576720177',
              label: 'REPORTS_TO',
              direction: 'TO',
              node: {},
            },
          },
        },
      ],
      return: ['*'],
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH nodes_cte AS (
  SELECT jsonb_agg(node) AS nodes
  FROM (
    -- Nodes from Employee
    SELECT jsonb_build_object(
      '_id', concat('Employee-', id_1691576718400.id::text),
      'label', 'Employee',
      'attributes', to_jsonb(id_1691576718400) - 'id'
    ) AS node
    FROM Employee id_1691576718400
  ) sub
),
edges_cte AS (
  SELECT '[]'::jsonb AS edges
),
node_counts AS (
  SELECT label, COUNT(*) AS count
  FROM (
    SELECT 'Employee' AS label FROM Employee
  ) t
  GROUP BY label
),
edge_counts AS (
    SELECT null AS label, 0 AS count
)
SELECT jsonb_build_object(
  'nodes', nodes_cte.nodes,
  'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
  'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM node_counts), '[]'::jsonb),
  'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM edge_counts), '[]'::jsonb)
) AS result
      WHERE id_1691576718400.title LIKE '%ale%'
      LIMIT 500;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });

  it('should return correctly on a query with both direction relation', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      limit: 500,
      logic: ['Like', '@id_1691576718400.title', '"ale"'],
      query: [
        {
          id: 'path_0',
          node: {
            id: 'id_1691576718400',
            label: 'Employee',
            relation: {
              id: 'id_1691576720177',
              label: 'REPORTS_TO',
              direction: 'BOTH',
              node: {},
            },
          },
        },
      ],
      return: ['*'],
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH nodes_cte AS (
  SELECT jsonb_agg(node) AS nodes
  FROM (
    -- Nodes from Employee
    SELECT jsonb_build_object(
      '_id', concat('Employee-', id_1691576718400.id::text),
      'label', 'Employee',
      'attributes', to_jsonb(id_1691576718400) - 'id'
    ) AS node
    FROM Employee id_1691576718400
  ) sub
),
edges_cte AS (
  SELECT '[]'::jsonb AS edges
),
node_counts AS (
  SELECT label, COUNT(*) AS count
  FROM (
    SELECT 'Employee' AS label FROM Employee
  ) t
  GROUP BY label
),
edge_counts AS (
    SELECT null AS label, 0 AS count
)
SELECT jsonb_build_object(
  'nodes', nodes_cte.nodes,
  'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
  'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM node_counts), '[]'::jsonb),
  'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM edge_counts), '[]'::jsonb)
) AS result
      WHERE id_1691576718400.title LIKE '%ale%'
      LIMIT 500;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });

  it('should return correctly on a query with relation logic', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      limit: 500,
      logic: ['<', '@id_1698231933579.unitPrice', '10'],
      query: [
        {
          id: 'path_0',
          node: {
            relation: {
              id: 'id_1698231933579',
              label: 'CONTAINS',
              depth: { min: 0, max: 1 },
              direction: 'TO',
              node: {},
            },
          },
        },
      ],
      return: ['*'],
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH nodes_cte AS (
  SELECT jsonb_agg(node) AS nodes
  FROM (
    -- Nodes from CONTAINS
    SELECT jsonb_build_object(
      '_id', concat('CONTAINS-', id_1698231933579.id::text),
      'label', 'CONTAINS',
      'attributes', to_jsonb(id_1698231933579) - 'id'
    ) AS node
    FROM CONTAINS id_1698231933579
  ) sub
),
edges_cte AS (
  SELECT '[]'::jsonb AS edges
),
node_counts AS (
  SELECT label, COUNT(*) AS count
  FROM (
    SELECT 'CONTAINS' AS label FROM CONTAINS
  ) t
  GROUP BY label
),
edge_counts AS (
    SELECT null AS label, 0 AS count
)
SELECT jsonb_build_object(
  'nodes', nodes_cte.nodes,
  'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
  'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM node_counts), '[]'::jsonb),
  'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM edge_counts), '[]'::jsonb)
) AS result
      WHERE id_1698231933579.unitPrice < 10
      LIMIT 500;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });

  it('should return correctly on a query with count logic', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      return: ['*'],
      logic: ['>', ['Count', '@p1'], '1'],
      query: [
        {
          id: 'path1',
          node: {
            label: 'Person',
            id: 'p1',
            relation: {
              label: 'DIRECTED',
              direction: 'TO',
              depth: { min: 1, max: 1 },
              node: {
                label: 'Movie',
                id: 'm1',
              },
            },
          },
        },
      ],
      limit: 5000,
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH nodes_cte AS (
  SELECT jsonb_agg(node) AS nodes
  FROM (
    -- Nodes from Person
    SELECT jsonb_build_object(
      '_id', concat('Person-', p1.id::text),
      'label', 'Person',
      'attributes', to_jsonb(p1) - 'id'
    ) AS node
    FROM Person p1
    UNION ALL
    -- Nodes from Movie
    SELECT jsonb_build_object(
      '_id', concat('Movie-', m1.id::text),
      'label', 'Movie',
      'attributes', to_jsonb(m1) - 'id'
    ) AS node
    FROM Movie m1
  ) sub
),
edges_cte AS (
  SELECT jsonb_agg(
    jsonb_build_object(
      '_id', concat('Person-', p1.id::text, '-Movie-', m1.id::text),
      'label', 'DIRECTED',
      'attributes', '{}'::jsonb,
      'from', p1.id::text,
      'to', m1.id::text
    )
  ) AS edges
  FROM Person p1
  JOIN Movie m1 ON p1.id = m1.id AND m1.relation = 'DIRECTED'
),
node_counts AS (
  SELECT label, COUNT(*) AS count
  FROM (
    SELECT 'Person' AS label FROM Person
    UNION ALL
    SELECT 'Movie' AS label FROM Movie
  ) t
  GROUP BY label
),
edge_counts AS (
  SELECT 'DIRECTED' AS label, COUNT(*) AS count FROM Person p1 JOIN Movie m1 ON p1.id = m1.id AND m1.relation = 'DIRECTED' GROUP BY label
)
SELECT jsonb_build_object(
  'nodes', nodes_cte.nodes,
  'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
  'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM node_counts), '[]'::jsonb),
  'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM edge_counts), '[]'::jsonb)
) AS result
      WHERE (SELECT COUNT(p1.id) FROM Person p1 JOIN Movie m1 ON p1.id = m1.id AND m1.relation = 'DIRECTED') > 1
      LIMIT 5000;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });

  it('should return correctly on a query with empty relation', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      limit: 500,
      query: [
        {
          id: 'path_0',
          node: {
            label: 'Movie',
            id: 'id_1730483610947',
            relation: {
              label: '',
              id: '',
              depth: { min: 0, max: 0 },
            },
          },
        },
      ],
      return: ['*'],
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH nodes_cte AS (
  SELECT jsonb_agg(node) AS nodes
  FROM (
    -- Nodes from Movie
    SELECT jsonb_build_object(
      '_id', concat('Movie-', id_1730483610947.id::text),
      'label', 'Movie',
      'attributes', to_jsonb(id_1730483610947) - 'id'
    ) AS node
    FROM Movie id_1730483610947
  ) sub
),
edges_cte AS (
  SELECT '[]'::jsonb AS edges
),
node_counts AS (
  SELECT label, COUNT(*) AS count
  FROM (
    SELECT 'Movie' AS label FROM Movie
  ) t
  GROUP BY label
),
edge_counts AS (
    SELECT null AS label, 0 AS count
)
SELECT jsonb_build_object(
  'nodes', nodes_cte.nodes,
  'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
  'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM node_counts), '[]'::jsonb),
  'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM edge_counts), '[]'::jsonb)
) AS result
      LIMIT 500;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });

  it('should return correctly on a query with upper case logic', () => {
    const query: BackendQueryFormat = {
      saveStateID: 'test',
      limit: 500,
      query: [
        {
          id: 'path_0',
          node: {
            id: 'id_1731428699410',
            label: 'Character',
          },
        },
      ],
      logic: ['Upper', '@id_1731428699410.name'],
      return: ['*'],
    };

    const sql = query2SQL(query);
    const expectedSQL = `WITH nodes_cte AS (
  SELECT jsonb_agg(node) AS nodes
  FROM (
    -- Nodes from Character
    SELECT jsonb_build_object(
      '_id', concat('Character-', id_1731428699410.id::text),
      'label', 'Character',
      'attributes', to_jsonb(id_1731428699410) - 'id'
    ) AS node
    FROM Character id_1731428699410
  ) sub
),
edges_cte AS (
  SELECT '[]'::jsonb AS edges
),
node_counts AS (
  SELECT label, COUNT(*) AS count
  FROM (
    SELECT 'Character' AS label FROM Character
  ) t
  GROUP BY label
),
edge_counts AS (
    SELECT null AS label, 0 AS count
)
SELECT jsonb_build_object(
  'nodes', nodes_cte.nodes,
  'edges', COALESCE(edges_cte.edges, '[]'::jsonb),
  'nodeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM node_counts), '[]'::jsonb),
  'edgeCounts', COALESCE((SELECT jsonb_agg(jsonb_build_object('label', label, 'count', count))
                           FROM edge_counts), '[]'::jsonb)
) AS result
      WHERE UPPER(id_1731428699410.name) IS NOT NULL
      LIMIT 500;`;

    expect(fixSQLSpaces(sql.query)).toBe(fixSQLSpaces(expectedSQL));
  });
});
