import { type DbConnection, type GraphQueryResultMetaFromBackend, graphQueryBackend2graphQuery } from 'ts-common';
import { Neo4jConnection } from 'ts-common/databaseConnection/neo4j';
import { log } from '../../logger';
import { QUERY_CACHE_DURATION, redis } from '../../variables';
import { parseCountCypherQueryResult, parseCypherQueryResult } from '../../queryExecution/cypher/queryResultParser';
import { cacheCheck } from './cache';
import type { QueryText } from '../../queryExecution/model';

export const cypherQueryService = async (
  db: DbConnection,
  cypher: QueryText,
  useCached: boolean,
): Promise<GraphQueryResultMetaFromBackend> => {
  let index = 0;
  const disambiguatedQuery = cypher.query.replace(/\d{13}/g, () => (index++).toString());
  const cacheKey = Bun.hash(JSON.stringify({ db: db, query: disambiguatedQuery })).toString();

  if (QUERY_CACHE_DURATION === '') {
    log.info('Query cache disabled, skipping cache check');
  } else if (!useCached) {
    log.info('Skipping cache check for query due to parameter', useCached);
  } else {
    const cachedMessage = await cacheCheck(cacheKey);
    if (cachedMessage) {
      log.debug('Cache hit for query', disambiguatedQuery);
      return cachedMessage;
    }
  }

  // TODO: only neo4j is supported for now
  const connection = new Neo4jConnection(db);
  try {
    const [neo4jResult, neo4jCountResult] = await connection.run([cypher.query, cypher.countQuery]);
    const graph = parseCypherQueryResult(neo4jResult.records);
    const countGraph = parseCountCypherQueryResult(neo4jCountResult.records);

    // calculate metadata
    const result = graphQueryBackend2graphQuery(graph, countGraph);
    result.nodeCounts.updatedAt = Date.now();

    // cache result
    const compressedMessage = Bun.gzipSync(JSON.stringify(result));
    const base64Message = Buffer.from(compressedMessage).toString('base64');

    if (QUERY_CACHE_DURATION !== '') {
      // if cache enabled, cache the result
      await redis.setWithExpire(cacheKey, base64Message, QUERY_CACHE_DURATION); // ttl in seconds
    }

    return result;
  } catch (error) {
    log.error('Error parsing query result:', cypher, error);
    throw new Error('Error parsing query result');
  } finally {
    connection.close();
  }
};
